@@ -20,10 +20,13 @@ function test()
   printBugNumber(BUGNUMBER);
   printStatus (summary);
 
-// ------- Comment #104 From Jesse Ruderman
-
-  (function(a) { function b() { a; } function a() { } })();
-// Assertion failure: pn_defn, at ../jsparse.h:655
+  try
+  {
+    delete (1 ? window : function(){});
+  }
+  catch(ex)
+  {
+  }
 
   reportCompare(expect, actual, summary);
 
